unit enumdirs;

{

  This unit is a modification of the NLDFileSearch unit which can be found at
  http://www.nldelphi.com/Forum
  
  Original authhor: Jos Visser aka GolezTrol (modifications by PsychoMark)
  Original authors website: http://www.goleztrol.nl
  
  Modification: Bart Broersma
  Website: http://home.tiscal.nl/~knmg0017
  
  
  This unit provides a procedure to recurse a directory, find all files
  in it, then use a callback procedure or event to perform some action on the specified file.
  
  Wildcard pattern matching for filenames is done by TExtMask.Matches().
  This is done for several reasons:
  - Linux does not provide this natively (like FindFirstFile() in win32)
  - If you set a mask in FindFirst/FindNext, the mask will also be applied to
    directorynames. This is (almost always) unwanted.
  - It also takes care of the problem that FindFirstFile() will also match the mask
    against the shortfilename (old DOS 8.3 format).
    For example: a file abcdefghij.txt has a shortname like abcdef~1.txt and will
    therefore (if we let Windows do the matching) match a mask like '*1.*', where infact
    it really should not. Taking care of pattern matching ourselfs solves this, since we only
    match long filenames
  - As a bonus it lets you use sets in a mask (like: [1..9]*.txt)
  At the current moment (Lazarus 0.9.25 SVN 14237) TMask.Matches() does it's matching
  in a case-insensitive way. This will of not always do for Linux, so I patched
  this for my own needs. This will be discussed with the Lazarus developers.
  
  
  Principal modofications to the original code of Golez Trol:
  - The main procedure originally was an event of a separate class, I did not like that
    so it's a normal procedure now. (This way it can be easily used in conventional non OOP based
    fpc programs)
  - The callbackroutine must supply a resultcode to indicate failure or success
  - Failure of the callback procedure can be ignored or will trigger an EEnumDirError exception
  - I overloaded the procedure so that it accepts either a mask (*.txt) or a masklist (*.txt;*.pp)
  - The procedure now also accepts a mask for exclusion
  - The original code also provided a recursive procedure to recurse the directory tree,
    I did not (yet?) included this. This procedure might cause stack overflows if the tree
    is too deep.
  - I separated the Path from the filespecs/wildcards. This is necessary for using MaskLists
    as opposed to a single mask. It also means that specifying filespecs or wildcards in the path
    will give wrong or unexpected results!!
  - Wildcard evaluation is done Linux style (on all platforms). You can alter this behaviour by setting
    the soMaskIsWindowsStyle option
  - Code is only tested on FPC/Lazarus. Passing procedures as a parameter to a procedure
    is done different on FPC then delphi. Therefore this code needs changing if it is ported for
    Delphi (remove the @) (I only have Delphi 3 and I cannot test this because this unit uses
    features in the Object Pascal language that D3 did not support back in those days)
  
  ToDo
  - Make FF/FN behave in a way that it does not follow symlinks (that is report the target
    instead of the actual symlink ('s name).
    This is currently discussed on the FreePascal bugtracker (http://bugs.freepascal.org)
  - Convince Lazarus developers that TMask.Matches needs patching so it can do
    case-sensitive pattern matching (as expected on Linux)
    
  History
  2008-01-16 First working version tested on Linux and Win32 platform
  2008-01-20 Added a TCustomDirEnumerator class that encapsulates the call to EnumDirEx()
             At the moment it only accepts a callback event, not a regular procedure.
  2008-01-23 Patched TMask.Matches() to do case-sensitive matching on Linux
  2008-01-03 Derived ExtMasks unit from Masks unit, so that the case-sensitivity of the pattern
             matching can be controlled with an extra argument for TExtMask.Create()
             Now you can do case-insensitive masking on linux too (easier if the mask only
             contains the extension (like *.jpg)
             I decided not to use compiler directives to force masks to be case-insensitive
             on Windows platform. It is the programmers responsibility to set the option.
             In multiplatform programs the programmer can ofcourse use compiler-directives to
             set soMaskIsWindowsStyle. In the end I decided this was the more flexible way.
  2008-02-24 Setting soMaskIsWindowsStyle now overrides soMaskIsCaseSensitive
             Bug fix: setting ErrorCode to FileProcess_User_Abort now exits the EnumDir() routine
             (by means of an exception) independent of the soIgnoreErrors option
             Bugfix: removed ExtractFilePath(Path), which truncated paths that did not end in
             PathDelimiter
  2008-12-25 Implemented an exclude masklist for the main procedure
  2008-03-12 Fixed MaskListWinToLinuxWildCards. It now correctly handles a mask like abc.*,
             which in Linux terms should both find abc and any file that begins with abc
             followed by a period (.).
  2008-03-31 New option soRecurseOnSymdirs, to allow/disallow recursing into directories
             pointed to by a symlink.
  


  If you use this code, either my original code, or modifications of it, credits must be given (either
  in the published source code, or in a routine in the program that display this information) to
  both me (Bart Broersma) and the authors of the original NLDFileSearch unit.

}


{$mode objfpc}{$H+}


interface

uses
  Classes, SysUtils; 
  
  
type
  TFileProcessorProc = Procedure(const FileName: string; const SR: TSearchRec; Out ErrorCode: LongInt);
  TFileProcessorEvent = Procedure(const FileName: string; const SR: TSearchRec; Out ErrorCode: LongInt) of Object;

  { SearchOptions
    soRecursive: Search the path recursively
    soProcessDirs: call the callbackroutine on directories also
    soDirsOnly: only call the callback routine on directories (ignore files)
    soProcessMessages: call Application.ProcessMessages after each call to FF/FN
      Yo can define a conditional define NOFORMS to disable the units depency on the Forms unit
    soUseMaskForDirs: use the searchmask for directories also.
      When set only directories whose name matches the mask will be reported and/or recursed.
      If set, this also applies to the ExclMaskList
    soFollowSymlinks: report the target of the symlink rather than the symlink itself
      The default behaviour of not following symlinks is not yet fully implemented due to
      the implementation of FF/FN in FreePascal (2.2.0)
    soRecurseOnSymDirs: recurse into directories even if it is in fact a symlink to a
      directory. Use with caution: this migth lead to an infinite loop if the symlink points
      to a directory that is higher in the current (e.g. current dir is /home/bart/test
      and symlink points to /home/bart). This option only has an effect if sorecursive is set
    soIgnoreErrors: continue recursing the directory tree if the callback routine reports an error
    soMaskIsCaseSensitive: matching the filename to the mask is done in a case sensitive way
      (so abc.txt does NOT match the mask ABC*)
      If set, this also applies to the ExclMaskList
    soMaskIsWindowsStyle: the given mask is based on Windows style matching. Default is Linux style.
      If you want a mask of '*.*' to match all files (that means even if they do not have an extension),
      you must set this option. Internally this means we translate the mask to the appropriate Linux
      style mask.
      If set, this also applies to the ExclMaskList
      Applying this option will disable the soCaseSensitive option, so masks in Windows style are
      treated as case insentive masks.
  }

  TfsOption = (soRecursive, soProcessDirs, soDirsOnly, soProcessMessages,
               soUseMaskForDirs, soFollowSymlinks, soRecurseOnSymDirs, soIgnoreErrors,
               soMaskIsCaseSensitive, soMaskIsWindowsStyle);
  TfsOptions = set of TfsOption;
     
  { EEnumDirError }

const
  MsgEnumDirError = 'An error ocurred in the filehandling procedure.';
  FileProcess_OK = 0; //A file handling routine should return this if all went well
  FileProcess_User_Abort = -2; //Return this if in your FileProcessor the user aborts the process
                               //I do not think any IO process throws this error code at you

type

  EEnumDirError = class(Exception)
    private
      fErrorCode: LongInt; //Returned code by filehandling routine
      fFileName: String; //Filename that was passed to filehandling routine
    public
      constructor Create(const Msg: String; const AErrorCode: LongInt; AFileName: String);
      property ErrorCode: LongInt read fErrorCode;
      property FileName: String read fFileName;
  end;


     
procedure EnumDir(Path: string; const Mask: string; const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorEvent);

procedure EnumDir(Path: string; const Mask: string; const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorProc);

procedure EnumDir(Path: string; const MaskList: string; ListSeparator: Char; const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorEvent);

procedure EnumDir(Path: string; const MaskList: string; ListSeparator: Char;  const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorProc);

//If you use an ExclMask, I only provide this as a MaskList, so you have to specify the ListSeparator!
procedure EnumDir(Path: string; const MaskList: string; const ExclMaskList: string; ListSeparator: Char; const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorEvent);

procedure EnumDir(Path: string; const MaskList: string; const ExclMaskList: string; ListSeparator: Char;  const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorProc);

procedure EnumDirEx(Path: string; MaskList: string; ExclMaskList: String; ListSeparator: Char;
                    Attr: LongInt; const SearchOptions: TfsOptions;
                    const FileHandlerProc: TFileProcessorProc;const FileHandlerEvent: TFileProcessorEvent);
                    
                    
type

  { TCustomDirEnumerator }

  TCustomDirEnumerator = class
  private
    fPath: string;
    fMaskList: string;
    fExclMaskList: string;
    fListSeparator: Char;
    fAttr: LongInt;
    fOptions: TfsOptions;
    fFileHandler: TFileProcessorEvent;
  public
    constructor Create;
    procedure EnumDir;
    property Path: string read fPath write fPath;
    property MaskList: string read fMaskList write fMaskList;
    property ExclMaskList: string read fExclMaskList write fExclMaskList;
    property ListSeparator: Char read fListSeparator write fListSeparator default ';';
    property Attribute: LongInt read fAttr write fAttr;
    property Options: TfsOptions read fOptions write fOptions;
    property FileHandler: TFileProcessorEvent read fFileHandler write fFileHandler;
  end;


  TDirEnumerator = class(TCustomDirEnumerator)
  published
    property Path;
    property MaskList;
    property ExclMaskList;
    property ListSeparator;
    property Attribute;
    property Options;
    property FileHandler;
  end;



procedure MaskListWinToLinuxWildCards(var AMaskList: String; const ASeparator: Char);
function IsLink(const Attr: LongInt): Boolean;




implementation

uses ExtMasks
     {$IFDEF DEBUG_ENUMDIRS},LCLProc {for DebugLn()}{$ENDIF}
     {$IFNDEF NOFORMS},Forms{$ENDIF};


function BoolStr(const Value: Boolean): String;
begin
  if Value then Result := 'True' else result := 'False';
end;

procedure MaskListWinToLinuxWildCards(var AMaskList: String; const ASeparator: Char);
//All we do here is change any mask that ends in '.*' so that it ends in '*'
//since in Linux if a mask ends in '.*', a '.' must be in the filename, whilst in Windows
//this translates to may or may not have any extension
var
  i: Integer;
  S: String;
  SL: TStringList;
begin
  (*
  if (ASeparator = #0) then
  begin
    exit;
  end;
  *)
  SL := TStringList.Create;
  Try
    SL.Delimiter := ASeparator;
    SL.DelimitedText := AMaskList;
    for i := 0 to SL.Count - 1 do
    begin
      S := SL.Strings[i];
      //A Windows mask that ends in a period ('.') is the same as the mask ending in '.*'
      if (S[Length(S)] = '.') then S := S + '*';
      //Wildcard end with '.*'  ?
      if (Length(S) > 1) and (S[Length(S)] = '*') and (S[Length(S)-1] = '.') then
      begin//Replace a wildcard ending in '.*' in one that ends in '*' if appropriate
        if (S = '.*') then S := '*'
        else
        begin//we know that Length(S) > 2 now...
          if (S[Length(S)-2] in ['?','*']) then
          begin
            System.Delete(S,Length(s),1);
            S[Length(S)] := '*';
            //Now the mask might end in '**' (if it ended in '*.*' before, so
            //if that is the case, we'll remove the last '*'
            if (Length(S) > 1) and (S[Length(S)-1] = '*') then System.Delete(S,Length(S),1);
          end
          else
          begin
            //at this point we know that S is something like abc.*
            //converting this to abc* is wrong, this would also find files like abcdef.*
            //we want to find in Linux terms both the file abc as well as abc.*
            //so we add the entry with the filename without the '.*'
            //and leave the current entry as it is
            SL.Add(System.Copy(S,1,Length(S)-2));
          end;
        end;//S <> '.*'
        SL.Strings[i]:= S;
      end;
    end;
    AMaskList := SL.DelimitedText;
  Finally
    SL.Free;
  end;
end;


function IsLink(const Attr: LongInt): Boolean;
{$IFDEF WINDOWS}
const FILE_ATTRIBUTE_REPARSE_POINT = $00000400;
{$ENDIF}
begin
  {$IFDEF MSWINDOWS}
  //ATM fpc (2.2.0) does not set faSymlink on Vista symlinks or NTFS ReparsePoints
  //so we need this dirty fix.
  Result := (Attr and (faSymlink or FILE_ATTRIBUTE_REPARSE_POINT)) > 0;
  {$ELSE}
  Result := (Attr and faSymlink) > 0;
  {$ENDIF}
end;

procedure EnumDirEx(Path: string; MaskList: string; ExclMaskList: String; ListSeparator: Char;
                    Attr: LongInt; const SearchOptions: TfsOptions;
                    const FileHandlerProc: TFileProcessorProc;const FileHandlerEvent: TFileProcessorEvent);
{
  - The caller must make sure that Path does NOT contain filespecs or wildards
  - If the caller specifies both a FilehandlerProc AND a FileHandlerEvent, then only the
    FileHandlerProc will be called as the callback routine
    (Normally the caller will use EnumDir() which in turn calls EnumDirEx() with only one of
    these set, so this should be no problem)
}
const Wild = '*';
var DosError: LongInt;
    DirList: TStringList;
    AMaskList, AExclMaskList: TExtMaskList;
    SR: TSearchRec;
    DoRecursive, DoProcessDirs, DoOnlyDirs,
    DoUseMaskForDirs, DoFollowSymlinks, DoRecurseOnSymDirs, DoIgnoreErrors,
    MaskIsCaseSensitive, MaskIsWindowsStyle, HasExclMask: Boolean;
    AbortFlag: Boolean;
    ErrorCode: LongInt;
    ErrFileName: String; //Used for EENumDirError
begin
  //This code must be compiled with boolean shortcut evaluation
  //otherwise it might call AExclMaskList.Matches() when AExclMaskList has not been created!!!
  {$B-}
  if (FileHandlerProc = nil) and (FileHandlerEvent = nil) then Exit;
  Path := ExpandFileName(Path);
  //This is WRONG!!
  //Path := ExtractFilePath(Path);
  //The Path must NOT contain filespecs or wildcards
  Path := IncludeTrailingPathDelimiter(Path);
  DoRecursive := (soRecursive in SearchOptions);
  DoOnlyDirs  := (soDirsOnly in SearchOptions); //This will set DoProcessDirs too
  DoProcessDirs := (soProcessDirs in SearchOptions) or DoOnlyDirs;
  DoUseMaskForDirs := (soUseMaskforDirs in SearchOptions);
  DoIgnoreErrors := (soIgnoreErrors in SearchOptions);
  DoFollowSymlinks := (soFollowSymlinks in SearchOptions);
  DoRecurseOnSymDirs := (soRecurseOnSymDirs in SearchOptions);
  MaskIsCaseSensitive := (soMaskIsCaseSensitive in SearchOptions);
  MaskIsWindowsStyle := (soMaskIsWindowsStyle in SearchOptions);
  HasExclMask := (Length(ExclMaskList) > 0) and (ExclMaskList <> #0);
  AbortFlag := False;
  ErrorCode := FileProcess_OK;
  ErrFileName := '';

  writeln('DoRecursive = '+BoolStr(DoRecursive));
  //DebugLn('DoOnlyDirs = '+BoolStr(DoOnlyDirs));
  //DebugLn('DoProcessDirs = '+BoolStr(DoProcessDirs));
  //DebugLn('DoUseMaskForDirs = '+BoolStr(DoUseMaskforDirs));
  //DebugLn('DoIgnoreErrors = '+BoolStr(DoIgnoreErrors));
  //DebugLn('DoFollowSymLinks = '+BoolStr(DoFollowSymlinks));
  writeln('MaskIsCaseSensitive = '+BoolStr(MaskIsCaseSensitive));
  writeln('Path = ',path);
  //DebugLn('MaskIsWindowsStyle = '+BoolStr(MaskIsWindowsStyle));

  //It is up to the programmer to set this option, I decided to not default this on DOS/Win
  if MaskIsWindowsStyle then
  begin
    //DebugLn('Masking is NOT case sensitive:');
    //DebugLn('  MaskList before conversion: ',MaskList);
    MaskIsCaseSensitive := False;
    MaskListWinToLinuxWildCards(MaskList, ListSeparator);
    if HasExclMask then MaskListWinToLinuxWildCards(ExclMaskList, ListSeparator);
    //DebugLn('  MaskList after conversion: ',MaskList);
  end;

  DirList := TStringList.Create;

  //Default behaviour for FF/FN is that mask is also applied to directories
  //This is most of the times unwanted for recursive searches
  //so we set up our own mask and try to match the filename to it
  //I preferr to create AMaskList once, instead of calling MatchesMaskList, which will each time
  //create a TExtMaskList object and free it after use
  AMaskList := TExtMaskList.Create(MaskList, MaskIsCaseSensitive, ListSeparator) ;
  if HasExclMask then AExclMaskList := TExtMaskList.Create(ExclMaskList, MaskIsCaseSensitive, ListSeparator)
  else AExclMaskList := nil;

  //Currently (FPC 2.2.0) FF/FN does follow symlinks by default in Linux (but not in win32)
  //In the bugtracker it is proposed that including faSymlink in Attr will (in future) make
  //FF/FN not follow symlinks, but report them as a symlink
  //If this will not be achieved, I will provide an alternative FF/FN for win32 and Linux platform
  //to be used with this unit.
  if not DoFollowSymlinks then Attr := Attr or faSymlink;
  try
    DirList.Add(Path);
    while (DirList.Count > 0) and (not AbortFlag) do
    begin
      //DebugLn('DirList.Count >0');
      Path := DirList[0];
      DirList.Delete(0);
      //DebugLn('Path = '+Path);
      //Find files in this dir
      //DebugLn('FindFirst('+Path + Wild +', $'+IntToHex(Attr, 2)+', SR)');
      
      DosError := FindFirst(Path + Wild, Attr, SR);
      //DebugLn('DosError = '+inttostr(doserror));
      while (DosError = 0) and (not AbortFlag)do
      begin
        //DebugLn('DosError = 0');
        if not ((SR.Name = '.') or (SR.Name = '..')) then
        begin//not . or ..
          if ((SR.Attr and faDirectory) > 0) then
          begin//It's a dir
            //We only handle this dir if we
            //1. do not use mask for dirs or we do use mask for dirs and the name matches the mask
            //2. do use mask for dirs and the dir does not match the exclude maks
            //DebugLn(SR.Name+' is a Directory');
            //DebugLn('  ((not DoUseMaskForDirs) or AMaskList.Matches(SR.Name)) = '+BoolStr(((not DoUseMaskForDirs) or AMaskList.Matches(SR.Name))));
            if (not DoUseMaskForDirs) or
               ((DoUseMaskForDirs and AMaskList.Matches(SR.Name)) and
                ((not HasExclMask) or (HasExclMask and not AExclMaskList.Matches(SR.Name)))
               ) then

            begin//Dirname matches
              //DebugLn('  Dirname matches');
              if DoRecursive then
              begin
                //if Dir is in fact a symlink to a dir we only recurse if soRecurseOnSymDirs is set
                if (not IsLink(SR.Attr)) or DoRecurseOnSymDirs then
                  DirList.Insert(0,Path + SR.Name + PathDelim);
              end;
              if DoProcessDirs then
              begin
                if Assigned(FileHandlerProc) then FileHandlerProc(Path+SR.Name, SR, ErrorCode)
                else if Assigned(FileHandlerEvent) then FileHandlerEvent(Path+SR.Name, SR, ErrorCode);
                //DebugLn('  ErrorCode = '+IntToStr(ErrorCode));
                if (ErrorCode <> FileProcess_OK) and
                   ((not DoIgnoreErrors) or (ErrorCode = FileProcess_User_Abort)) then AbortFlag := True;
                //DebugLn('  AbortFlag = '+BoolStr(AbortFlag));
              end;
            end;//Dirname Matches
          end//It's a dir
          else
          begin//It's a file
            //We handle this file only if we
            //1. do not only handle dirs
            //2. the filename matches the mask
            //3. the filename does not match the excludemask or we do not have an exclude mask
            //DebugLn(SR.Name +' is a file');
            //DebugLn('  (not DoOnlyDirs) = '+BoolStr(not DoOnlyDirs)+', AMaskList.Matches('+SR.Name+') = '+BoolStr(AMaskList.Matches(SR.Name)));
            if (not DoOnlyDirs) and AMaskList.Matches(SR.Name) and
               ((not HasExclMask) or (not AExclMaskList.Matches(SR.Name))) then
            begin//Process file
              //DebugLn('  Processing file');
              if Assigned(FileHandlerProc) then FileHandlerProc(Path+SR.Name, SR, ErrorCode)
              else if Assigned(FileHandlerEvent) then FileHandlerEvent(Path+SR.Name, SR, ErrorCode);
              //DebugLn('  ErrorCode = '+IntToStr(ErrorCode));
              if (ErrorCode <> FileProcess_OK) and
                 ((not DoIgnoreErrors) or (ErrorCode = FileProcess_User_Abort)) then AbortFlag := True;
              //DebugLn('  AbortFlag = '+BoolStr(AbortFlag));
            end;//Process file
          end;//It's a file
          if AbortFlag then ErrFileName := Path + SR.Name;
          {$IFNDEF NOFORMS}
          if (soProcessMessages in SearchOptions) then Application.ProcessMessages;
          {$ENDIF}
        end;//not . or ..
        DosError := FindNext(SR);
        //DebugLn('FindNext(SR) = '+IntToStr(DosError));
      end;//DosError <> 0
      FindClose(SR);
    end;//DirList.Count > 0
    if AbortFlag then Raise EEnumDirError.Create(MsgEnumDirError,ErrorCode, ErrFileName);
  finally
    DirList.Free;
    AMaskList.Free;
    if HasExclMask then AExclMaskList.Free;
  end;
end;


procedure EnumDir(Path: string; const Mask: string; const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorEvent);
begin
  EnumDirEx(Path, Mask, '', ';', Attr, SearchOptions, nil, FileHandler);
end;

procedure EnumDir(Path: string; const Mask: string; const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorProc);
begin
  EnumDirEx(Path, Mask, '', ';', Attr, SearchOptions, FileHandler, nil);
end;

procedure EnumDir(Path: string; const MaskList: string; ListSeparator: Char; const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorEvent);
begin
  EnumDirEx(Path, MaskList, '', ListSeparator, Attr, SearchOptions, nil, FileHandler);
end;

procedure EnumDir(Path: string; const MaskList: string; ListSeparator: Char;  const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorProc);
begin
  EnumDirEx(Path, MaskList, '', ListSeparator, Attr, SearchOptions, FileHandler, nil);
end;



procedure EnumDir(Path: string; const MaskList: string; const ExclMaskList: string; ListSeparator: Char; const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorEvent);
begin
  EnumDirEx(Path, MaskList, ExclMaskList, ListSeparator, Attr, SearchOptions, nil, FileHandler);
end;

procedure EnumDir(Path: string; const MaskList: string; const ExclMaskList: string; ListSeparator: Char;  const Attr: LongInt;
                  const SearchOptions: TfsOptions; const FileHandler: TFileProcessorProc);
begin
  EnumDirEx(Path, MaskList, ExclMaskList, ListSeparator, Attr, SearchOptions, FileHandler, nil);
end;



{ EEnumDirError }

constructor EEnumDirError.Create(const Msg: String; const AErrorCode: LongInt; AFileName: String);
begin
  inherited Create(Msg);
  fErrorCode := AErrorCode;
  fFileName := AFileName;
end;

{ TCustomDirEnumerator }

constructor TCustomDirEnumerator.Create;
begin
  fListSeparator := ';';
  fExclMaskList := '';
end;

procedure TCustomDirEnumerator.EnumDir;
begin
  if Assigned(fFileHandler) then
    EnumDirEx(fPath, fMaskList, fExclMaskList, fListSeparator, fAttr, fOptions, nil, fFileHandler);
end;





{ from /usr/share/fpcrsc/rtl/unix/unixutil.pp
  This function matches filenames Linux style, not only wildcards, but also
  case-sensitive, in contrast to TExtMask.Matches() which is case insensitive
  
  However it does not match wildcards like [a..d], which TExtMask.Matches does
}


{

Function FNMatch(const Pattern,Name:string):Boolean;
Var
  LenPat,LenName : longint;

  Function DoFNMatch(i,j:longint):Boolean;
  Var
    Found : boolean;
  Begin
  Found:=true;
  While Found and (i<=LenPat) Do
   Begin
     Case Pattern[i] of
      '?' : Found:=(j<=LenName);
      '*' : Begin
            //find the next character in pattern, different of ? and *
              while Found do
                begin
                inc(i);
                if i>LenPat then Break;
                case Pattern[i] of
                  '*' : ;
                  '?' : begin
                          if j>LenName then begin DoFNMatch:=false; Exit; end;
                          inc(j);
                        end;
                else
                  Found:=false;
                end;
               end;
              Assert((i>LenPat) or ( (Pattern[i]<>'*') and (Pattern[i]<>'?') ));
            //Now, find in name the character which i points to, if the * or
            // wasn't the last character in the pattern, else, use up all the
            // chars in name
              Found:=false;
              if (i<=LenPat) then
              begin
                repeat
                  //find a letter (not only first !) which maches pattern[i]
                  while (j<=LenName) and (name[j]<>pattern[i]) do
                    inc (j);
                  if (j<LenName) then
                  begin
                    if DoFnMatch(i+1,j+1) then
                    begin
                      i:=LenPat;
                      j:=LenName;//we can stop
                      Found:=true;
                      Break;
                    end else
                      inc(j);//We didn't find one, need to look further
                  end else
                  if j=LenName then
                  begin
                    Found:=true;
                    Break;
                  end;
                  // This 'until' condition must be j>LenName, not j>=LenName.
                  //  That's because when we 'need to look further' and
                  //  j = LenName then loop must not terminate.
                until (j>LenName);
              end else
              begin
                j:=LenName;//we can stop
                Found:=true;
              end;
            end;
     else //not a wildcard character in pattern
       Found:=(j<=LenName) and (pattern[i]=name[j]);
     end;
     inc(i);
     inc(j);
   end;
  DoFnMatch:=Found and (j>LenName);
  end;

Begin //start FNMatch
  LenPat:=Length(Pattern);
  LenName:=Length(Name);
  FNMatch:=DoFNMatch(1,1);
End;









}












end.


